import builtins
from collections import defaultdict, OrderedDict
from threading import Lock
from typing import Dict, Any, Sequence, List

from .semantics.analyzer import NodeParents
from .thunk import ValueThunk, NodeThunk, Thunk
from .statements import GlobalStatement, GlobalImport, GlobalImportFrom, Definitions
from .exceptions import EntryError, SemanticError

ScopeDict = Dict[str, GlobalStatement]


class Builtins(dict):
    def __init__(self, injections: dict):
        base = dict(vars(builtins))
        common = set(base) & set(injections)
        if common:
            raise SemanticError('Some injections clash with builtins: ' + str(common))
        base.update(injections)
        super().__init__(base)

    def __getitem__(self, name):
        try:
            return super().__getitem__(name)
        except KeyError:
            raise EntryError(f'"{name}" is not defined.') from None


class Scope:
    """
    A lazy global scope.

    Internal logic:
        - each name is bound to a single statement
        - each statement has a target that can be rendered
        - multiple targets can map to the same statement: `a = b = f(...)`
    """

    def __init__(self, definitions: Definitions, parent, parents: NodeParents):
        super().__init__()
        self.definitions = definitions
        self._parent = parent

        self._name_to_thunk: OrderedDict[str, Thunk] = OrderedDict()
        self._statement_to_names: Dict[GlobalStatement, List[str]] = {}
        self._parents = parents

        self._populated = False
        self._updated = False

        locks = {}
        for definition in definitions:
            statement = definition.statement
            if statement not in locks:
                locks[statement] = Lock()

            self._name_to_thunk[definition.name] = NodeThunk(statement, locks[statement])

        self._update_names()

    def __getitem__(self, name: str):
        """ Render a name """
        if name not in self._name_to_thunk:
            return self._parent[name]

        thunk = self._name_to_thunk[name]
        if thunk.ready:
            return thunk.value

        assert isinstance(thunk, NodeThunk)
        with thunk.lock:
            if not thunk.ready:
                self._populated = True

                statement = thunk.statement
                names = self._statement_to_names[statement]

                for name, value in zip(names, statement.render(self, names)):
                    local = self._name_to_thunk[name]
                    assert isinstance(local, NodeThunk)
                    assert local.lock == thunk.lock
                    local.set(value)

            return thunk.value

    def __contains__(self, name):
        return name in self._name_to_thunk or name in self._parent

    def check_populated(self):
        if self._populated:
            raise RuntimeError('The scope has already been populated with live objects. Overwriting them might cause '
                               'undefined behaviour. Please, create another instance of Config.')

    def keys(self):
        return self._name_to_thunk.keys()

    def update_values(self, values: Dict[str, Any]):
        self.check_populated()
        self._updated = True

        for name, value in values.items():
            self._name_to_thunk[name] = ValueThunk(value)

        self._update_names()

    def _update_names(self):
        result = defaultdict(list)
        for name, thunk in self._name_to_thunk.items():
            if isinstance(thunk, NodeThunk):
                result[thunk.statement].append(name)
        self._statement_to_names = dict(result)

    def _get_leave_time(self, entry_points: Sequence[str]):
        def mark_name(name):
            nonlocal current
            if name not in leave_time:
                leave_time[name] = current
                current += 1

        def visit_parents(name):
            visited.add(name)
            for parent in self._parents[name]:
                find_leave_time(parent)

        def find_leave_time(name):
            if name in visited:
                return

            visit_parents(name)
            mark_name(name)

        names = {d.name: d for d in self.definitions}
        if entry_points is None:
            entry_points = list(names)
        else:
            delta = set(entry_points) - set(names)
            if delta:
                raise ValueError('The names %s are not defined, and cannot be used as entry points.' % delta)

        leave_time = {}
        visited = set()
        current = 0
        # we can't just visit the first-level nodes because some of them may have several names
        #  we need to drop such cases
        for n in entry_points:
            visit_parents(n)
            mark_name(n)

        return [names[n] for n in leave_time], {names[n]: t for n, t in leave_time.items()}

    def render(self, entry_points: Sequence[str] = None):
        if self._updated:
            raise RuntimeError('The scope has already been updated by live objects that cannot be rendered properly.')

        definitions, order = self._get_leave_time(entry_points)
        return render_scope(definitions, order)

    def __setitem__(self, key, value):
        raise NotImplementedError


def render_scope(definitions, order):
    # imports
    import_groups, imports, rest = defaultdict(list), [], []
    for definition in sorted(definitions, key=lambda s: order[s]):
        statement = definition.statement
        if isinstance(statement, GlobalImportFrom):
            import_groups[statement.root].append(statement)
        elif isinstance(statement, GlobalImport):
            imports.append(statement)
        else:
            rest.append(definition)

    for statement in imports:
        yield statement.to_str()

    if imports:
        yield ''

    for group in import_groups.values():
        yield GlobalImportFrom.group_to_str(group)

    if import_groups or imports:
        yield '\n'

    # definitions
    groups = defaultdict(list)
    for definition in rest:
        groups[definition.statement].append(definition)

    for _, group in sorted(groups.items(), key=lambda x: min(order[s] for s in x[1])):
        assert len(set(map(type, group))) == 1
        yield group[0].statement.group_to_str(group)



import ast

from .visitor import SemanticVisitor
from ..parser import extract_assign_targets


class LocalsGatherer(SemanticVisitor):
    def __init__(self):
        self.names = []

    @classmethod
    def gather(cls, nodes):
        instance = cls()
        instance._iterate_nodes(nodes)
        return instance.names

    def visit_function_def(self, node: ast.FunctionDef):
        self.names.append(node.name)

    def visit_assign(self, node: ast.Assign):
        self.names.extend(extract_assign_targets(node.targets))

    def generic_visit(self, node: ast.AST, *args, **kwargs):
        pass



import argparse
from pathlib import Path

from lazycon import load


def run(config, expression):
    value = load(config).eval(expression)
    if value is not None:
        print(value)


def build(input, output, entry_points):
    output = Path(output)
    output.parent.mkdir(parents=True, exist_ok=True)
    load(input).dump(output, entry_points)


def main():
    parser = argparse.ArgumentParser(description='Shortcuts for operations with config files.')
    subparsers = parser.add_subparsers()

    new = subparsers.add_parser('run', help='Run the given `config` by evaluating the passed `expression`.')
    new.set_defaults(callback=run)
    new.add_argument('config', help='path to a config file.')
    new.add_argument('expression', help='the entry point.')

    new = subparsers.add_parser(
        'build', help='Read the `input` config and flatten all its imports in order to obtain '
                      'an `output` config without dependencies.')
    new.set_defaults(callback=build)
    new.add_argument('input', help='path to the input config file.')
    new.add_argument('output', help='path to the output config file.')
    new.add_argument('-ep', '--entry_points', nargs='+', default=None,
                     help='names that should be kept during rendering.')

    args = vars(parser.parse_args())
    if 'callback' not in args:
        parser.print_help()
    else:
        callback = args.pop('callback')
        callback(**args)



import ast
from collections import defaultdict
from enum import Enum
from typing import Iterable, List, Dict, Optional

from .locals import LocalsGatherer
from ..parser import IGNORE_NAME, extract_assign_targets
from ..statements import GlobalStatement, GlobalFunction, GlobalAssign, GlobalImport, Definitions
from ..exceptions import SemanticError
from .visitor import SemanticVisitor


def position(node: ast.AST):
    return node.lineno, node.col_offset


# TODO: __folder__
READ_ONLY = {'__file__'}
NodeParents = Dict[str, List[str]]


class Semantics(SemanticVisitor):
    def __init__(self, definitions: Definitions, builtins: Iterable[str]):
        self.messages = defaultdict(lambda: defaultdict(set))

        # scopes
        self._builtins = builtins
        self._global_scope: Dict[str, MarkedValue] = {}
        self._statement_names: Dict[GlobalStatement, List[str]] = defaultdict(list)
        self._local_scopes: List[Dict[str, Marked]] = []

        marks = {}
        for definition in definitions:
            self._global_scope[definition.name] = marks.setdefault(
                definition.statement, MarkedValue(definition.statement))
            self._statement_names[definition.statement].append(definition.name)

            *pos, source = definition.statement.position

            if definition.name in READ_ONLY:
                self.add_message('The value is read-only', f'"{definition.name}" at %d:%d' % tuple(pos), source)

        self._statement_names = dict(self._statement_names)

        # tracking
        self._global_statement: Optional[GlobalStatement] = None
        # TODO: use ordered set
        self.parents: NodeParents = {d.name: [] for d in definitions}

        # analysis
        for statement in marks:
            self.visit(statement)

    @staticmethod
    def format(message, elements):
        message += ':\n'
        for source, item in elements.items():
            message += '  in %s\n    ' % source
            message += ', '.join(item)
            message += '\n'
        return message

    def check(self):
        message = ''
        for msg, elements in self.messages.items():
            message += self.format(msg, elements)
        if message:
            raise SemanticError(message)

    def add_message(self, message, content, source=None):
        # TODO: move line info here?
        source = source or self._global_statement.source_path
        self.messages[message][source].add(content)

    # scope management

    def enter_scope(self, names: Iterable[str], visited: Iterable[str] = ()):
        scope = {}
        for name in names:
            scope[name] = Marked(VisitState.Undefined)
        for name in visited:
            scope[name] = Marked(VisitState.Defined)
        self._local_scopes.append(scope)

    def leave_scope(self):
        self._local_scopes.pop()

    def enter(self, name: str):
        if self._local_scopes:
            value = self._local_scopes[-1][name]
            # allow multiple definitions
            if value.state is not VisitState.Defined:
                value.enter()

        else:
            self._global_scope[name].enter()

    def leave(self, name: str):
        if self._local_scopes:
            value = self._local_scopes[-1][name]
            # allow multiple definitions
            if value.state is not VisitState.Defined:
                value.leave()

        else:
            self._global_scope[name].leave()

    def generic_visit(self, node: ast.AST, *args, **kwargs):
        self.add_message('This syntactic structure is not supported',
                         f'{type(node).__name__} at %d:%d' % position(node))

    # the most important part - variable resolving

    def visit_name(self, node: ast.Name):
        assert isinstance(node.ctx, ast.Load), node.ctx
        name = node.id
        if name == IGNORE_NAME:
            self.add_message(f'The name "{IGNORE_NAME}" can only be used as wildcard during unpacking',
                             'at %d:%d' % position(node))

        # local scopes
        for level, scope in enumerate(reversed(self._local_scopes)):
            if name in scope:
                value = scope[name]
                # allow late binding
                if level == 0 and value.state is not VisitState.Defined:
                    self.add_message('Local variables referenced before being defined',
                                     '"' + name + '" at %d:%d' % position(node))
                return

        # global scope
        if name in self._global_scope:
            value = self._global_scope[name]
            if value.state is VisitState.Defining:
                self.add_message('Values are referenced before being completely defined (cyclic dependency)',
                                 '"' + name + '" at %d:%d' % position(node))

            for current in self._statement_names[self._global_statement]:
                self.parents[current].append(name)
            return

        # builtins
        if name not in self._builtins:
            self.add_message('Undefined names found', name)

    # global definitions

    def visit_global_assign(self, statement: GlobalAssign):
        assert self._global_statement is None
        self._global_statement = statement

        # we can just pick the first name - the rest will enter the same state automatically
        self.enter(self._statement_names[statement][0])
        self.visit(statement.node.value)
        self.leave(self._statement_names[statement][0])

        self._global_statement = None

    def visit_global_function(self, statement: GlobalFunction):
        assert self._global_statement is None
        self._global_statement = statement

        self.visit(statement.node)

        self._global_statement = None

    def visit_global_import(self, statement: GlobalImport):
        self.enter(statement.name)
        self.leave(statement.name)

    visit_global_import_from = visit_global_import

    # local definitions

    def visit_assign(self, node: ast.Assign):
        names = extract_assign_targets(node.targets)
        for name in names:
            self.enter(name)

        self.visit(node.value)

        for name in names:
            self.leave(name)

    def visit_function_def(self, node: ast.FunctionDef):
        self.enter(node.name)
        self.visit(node.args)
        # TODO: type annotations?
        self._iterate_nodes(node.decorator_list)
        self.leave(node.name)

        # ignore docstring
        body = node.body
        if isinstance(body[0], ast.Expr) and isinstance(body[0].value, ast.Str):
            body = body[1:]

        self.enter_scope(LocalsGatherer.gather(body), self._gather_arg_names(node.args))
        self._iterate_nodes(body)
        self.leave_scope()

    # other stuff that manages scope

    def visit_lambda(self, node: ast.Lambda):
        self.visit(node.args)

        self.enter_scope([], self._gather_arg_names(node.args))
        self.visit(node.body)
        self.leave_scope()

    def visit_list_comp(self, node):
        for comp in node.generators:
            self.visit(comp)

        self.visit(node.elt)

        for _ in node.generators:
            self.leave_scope()

    def visit_dict_comp(self, node):
        for comp in node.generators:
            self.visit(comp)

        self.visit(node.key)
        self.visit(node.value)

        for _ in node.generators:
            self.leave_scope()

    visit_set_comp = visit_generator_exp = visit_list_comp

    def visit_comprehension(self, node: ast.comprehension):
        assert not getattr(node, 'is_async', False)

        def get_names(target):
            assert isinstance(target.ctx, ast.Store)
            if isinstance(target, (ast.Tuple, ast.List)):
                names = []
                for elt in target.elts:
                    names.extend(get_names(elt))
                return names

            if isinstance(target, ast.Starred):
                return [target.value]

            assert isinstance(target, ast.Name), target
            return [target.id]

        self.visit(node.iter)
        self.enter_scope({}, get_names(node.target))

        for test in node.ifs:
            self.visit(test)

    # function-related stuff

    def visit_return(self, node: ast.Return):
        self._visit_valid(node.value)

    @staticmethod
    def _gather_arg_names(node: ast.arguments):
        args = getattr(node, 'posonlyargs', []) + node.args + node.kwonlyargs
        if node.vararg is not None:
            args.append(node.vararg)
        if node.kwarg is not None:
            args.append(node.kwarg)
        return [arg.arg for arg in args]


class VisitState(Enum):
    Undefined, Defining, Defined = 0, 1, 2


class Marked:
    def __init__(self, status: VisitState):
        self.state = status

    def enter(self):
        assert self.state is VisitState.Undefined, self.state
        self.state = VisitState.Defining

    def leave(self):
        assert self.state is not VisitState.Defined, self.state
        self.state = VisitState.Defined


class MarkedValue(Marked):
    def __init__(self, value):
        super().__init__(VisitState.Undefined)
        self.value = value



import ast
import os
import sys
from typing import Sequence, NamedTuple, Any, Tuple

from .exceptions import ConfigImportError
from .render import ScopeExec, execute

IGNORE_NAME = '_'


class Wrapper:
    def __init__(self, position):
        self.line, self.column, self.source_path = self.position = position


class GlobalStatement(Wrapper):
    """
    Wraps ast nodes used to define names in configs and maps a single name to a ast node.
    """

    def __init__(self, node: ast.AST, position):
        super().__init__(position)
        self.node = node

    def render(self, global_scope, names: Sequence[str]) -> Sequence[Any]:
        wrapper = ScopeExec(global_scope)
        execute(self.node, wrapper, self.source_path)
        return wrapper.get_results(names)


class Definition(NamedTuple):
    name: str
    statement: GlobalStatement


Definitions = Sequence[Definition]


class GlobalAssign(GlobalStatement):
    """ A single `target = value` statement. """

    def __init__(self, node: ast.Assign, body: str, position):
        super().__init__(node, position)
        self.body = body

    @staticmethod
    def _group_to_str(pattern, names, level) -> Tuple[str, bool]:
        if isinstance(pattern, ast.Name):
            name = pattern.id if pattern.id in names else IGNORE_NAME
            # the name could already have been IGNORE_NAME
            return name, name != IGNORE_NAME

        if isinstance(pattern, ast.Starred):
            child, contains = GlobalAssign._group_to_str(pattern.value, names, level + 1)
            return '*' + child, contains

        assert isinstance(pattern, (ast.Tuple, ast.List))
        joined, contains = [], False
        for elt in pattern.elts:
            child, local = GlobalAssign._group_to_str(elt, names, level + 1)
            contains = contains or local
            joined.append(child)
        joined = ', '.join(joined)

        if isinstance(pattern, ast.Tuple):
            if len(pattern.elts) == 1:
                joined += ','
            if level > 0:
                joined = f'({joined})'

        else:
            joined = f'[{joined}]'

        return joined, contains

    @staticmethod
    def group_to_str(definitions: Definitions):
        assert len({d.statement for d in definitions}) == 1
        statement: GlobalAssign = definitions[0].statement
        names = {d.name for d in definitions}

        result = ''
        for target in statement.node.targets:
            target, contains = GlobalAssign._group_to_str(target, names, 0)
            if contains:
                result += target + ' = '

        return result + statement.body


class GlobalFunction(GlobalStatement):
    def __init__(self, node: ast.FunctionDef, body: str, position):
        super().__init__(node, position)
        self.body = body

    def to_str(self):
        return '\n' + self.body.strip() + '\n\n'

    @staticmethod
    def group_to_str(definitions: Definitions):
        assert len(definitions) == 1
        return definitions[0].statement.to_str()

    def render(self, global_scope, names: Sequence[str]) -> Sequence[Any]:
        assert len(names) == 1 and self.node.name == names[0]
        return super().render(global_scope, names)


class GlobalImport(GlobalStatement):
    def __init__(self, node: ast.Import, position):
        super().__init__(node, position)
        alias, = node.names
        self.alias = alias
        self.name = alias.asname or alias.name.split('.', 1)[0]

    def _import_what(self):
        alias = self.alias
        result = alias.name
        if alias.asname is not None:
            result += ' as ' + alias.asname

        return result

    def to_str(self):
        return f'import {self._import_what()}'

    def render(self, global_scope, names: Sequence[str]) -> Sequence[Any]:
        assert len(names) == 1 and self.name == names[0]
        return super().render(global_scope, names)


class GlobalImportFrom(GlobalImport):
    def __init__(self, node: ast.ImportFrom, position):
        super().__init__(node, position)
        self.root = node.module

    def to_str(self):
        return f'from {self.root} {super().to_str()}'

    @staticmethod
    def group_to_str(statements: Sequence['GlobalImportFrom']):
        base, *statements = statements
        result = base.to_str()
        for statement in statements:
            result += ', ' + statement._import_what()

        return result


class ImportConfig(Wrapper):
    def __init__(self, root: Sequence[str], dots: int, extension: str, position):
        super().__init__(position)
        self.dots = dots
        self.root = tuple(root)
        self.position = position
        self.extension = extension

    def get_path(self, shortcuts):
        # relative import
        if self.dots > 0:
            root = (os.pardir,) * (self.dots - 1) + self.root
            prefix = os.path.dirname(self.source_path)
            return os.path.join(prefix, *root) + self.extension

        # import by shortcut
        shortcut, *root = self.root
        if shortcut in shortcuts:
            return os.path.join(shortcuts[shortcut], *root) + self.extension

        # import by sys.path
        visited = set()
        for prefix in sys.path:
            # optimizing disk access
            if prefix in visited:
                continue
            visited.add(prefix)

            path = os.path.join(prefix, *self.root) + self.extension
            if os.path.exists(path):
                return path

        root = '.'.join(self.root)
        raise ConfigImportError(f'Parent config "{root}" not found while parsing "{self.source_path}".')



import ast
import re
from typing import Iterable

first_cap = re.compile('(.)([A-Z][a-z]+)')
all_cap = re.compile('([a-z0-9])([A-Z])')


def snake_case(name):
    name = first_cap.sub(r'\1_\2', name)
    return all_cap.sub(r'\1_\2', name).lower()


class Visitor:
    def visit(self, node: ast.AST, *args, **kwargs):
        method = 'visit_' + snake_case(node.__class__.__name__)
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node, *args, **kwargs)

    def generic_visit(self, node: ast.AST, *args, **kwargs):
        raise NotImplementedError(node.__class__.__name__)

    def _iterate_nodes(self, nodes: Iterable, *args, **kwargs):
        for item in nodes:
            self.visit(item, *args, **kwargs)

    def _visit_nodes(self, nodes: Iterable, *args, **kwargs):
        result = []
        for item in nodes:
            result.append(self.visit(item, *args, **kwargs))
        return result
